from numbers import Real

from . import bounds
from .hints import Point
from .utils import (sum_expansions,
                    to_cross_product,
                    two_diff_tail,
                    two_product,
                    two_two_diff)


def signed_area(first_start: Point, first_end: Point,
                second_start: Point, second_end: Point) -> Real:
    """
    Calculates signed area of parallelogram built on segments' vectors.

    Positive sign of result means that second vector is counterclockwise,
    negative -- clockwise,
    zero -- collinear to first vector.

    >>> signed_area((0, 0), (1, 0), (0, 0), (1, 0))
    0
    >>> signed_area((0, 0), (1, 0), (0, 0), (0, 1))
    1
    >>> signed_area((0, 0), (1, 0), (0, 1), (0, 0))
    -1
    """
    first_start_x, first_start_y = first_start
    first_end_x, first_end_y = first_end
    second_start_x, second_start_y = second_start
    second_end_x, second_end_y = second_end

    minuend = ((first_end_x - first_start_x)
               * (second_end_y - second_start_y))
    subtrahend = ((first_end_y - first_start_y)
                  * (second_end_x - second_start_x))
    result = minuend - subtrahend

    if minuend > 0:
        if subtrahend <= 0:
            return result
        else:
            upper_bound = minuend + subtrahend
    elif minuend < 0.0:
        if subtrahend >= 0.0:
            return result
        else:
            upper_bound = -minuend - subtrahend
    else:
        return result

    error_bound = bounds.to_signed_measure_first_error(upper_bound)
    if result >= error_bound or -result >= error_bound:
        return result

    return _adjusted_signed_area(first_start_x, first_start_y,
                                 first_end_x, first_end_y,
                                 second_start_x, second_start_y,
                                 second_end_x, second_end_y,
                                 upper_bound)


def _adjusted_signed_area(first_start_x: Real, first_start_y: Real,
                          first_end_x: Real, first_end_y: Real,
                          second_start_x: Real, second_start_y: Real,
                          second_end_x: Real, second_end_y: Real,
                          upper_bound: Real) -> Real:
    minuend_multiplier_x = first_end_x - first_start_x
    minuend_multiplier_y = second_end_y - second_start_y
    subtrahend_multiplier_x = second_end_x - second_start_x
    subtrahend_multiplier_y = first_end_y - first_start_y

    minuend_tail, minuend = two_product(minuend_multiplier_x,
                                        minuend_multiplier_y)
    subtrahend_tail, subtrahend = two_product(subtrahend_multiplier_y,
                                              subtrahend_multiplier_x)

    result_expansion = two_two_diff(minuend_tail, minuend, subtrahend_tail,
                                    subtrahend)
    result = sum(result_expansion)
    error_bound = bounds.to_signed_measure_second_error(upper_bound)
    if result >= error_bound or -result >= error_bound:
        return result

    minuend_multiplier_x_tail = two_diff_tail(first_end_x, first_start_x,
                                              minuend_multiplier_x)
    subtrahend_multiplier_x_tail = two_diff_tail(second_end_x,
                                                 second_start_x,
                                                 subtrahend_multiplier_x)
    subtrahend_multiplier_y_tail = two_diff_tail(first_end_y, first_start_y,
                                                 subtrahend_multiplier_y)
    minuend_multiplier_y_tail = two_diff_tail(second_end_y, second_start_y,
                                              minuend_multiplier_y)
    if (not minuend_multiplier_x_tail
            and not minuend_multiplier_y_tail
            and not subtrahend_multiplier_x_tail
            and not subtrahend_multiplier_y_tail):
        return result

    error_bound = (bounds.to_signed_measure_third_error(upper_bound)
                   + bounds.to_determinant_error(result))
    result += ((minuend_multiplier_x * minuend_multiplier_y_tail
                + minuend_multiplier_y * minuend_multiplier_x_tail)
               - (subtrahend_multiplier_y * subtrahend_multiplier_x_tail
                  + subtrahend_multiplier_x * subtrahend_multiplier_y_tail))
    if result >= error_bound or -result >= error_bound:
        return result

    result_expansion = sum_expansions(
            result_expansion, to_cross_product(minuend_multiplier_x_tail,
                                               minuend_multiplier_y,
                                               subtrahend_multiplier_x,
                                               subtrahend_multiplier_y_tail))
    result_expansion = sum_expansions(
            result_expansion, to_cross_product(minuend_multiplier_x,
                                               minuend_multiplier_y_tail,
                                               subtrahend_multiplier_x_tail,
                                               subtrahend_multiplier_y))
    result_expansion = sum_expansions(
            result_expansion, to_cross_product(minuend_multiplier_x_tail,
                                               minuend_multiplier_y_tail,
                                               subtrahend_multiplier_x_tail,
                                               subtrahend_multiplier_y_tail))
    return result_expansion[-1]



from numbers import Real

from . import parallelogram
from .hints import Point
from .utils import to_perpendicular_point


def signed_length(first_start: Point, first_end: Point,
                  second_start: Point, second_end: Point) -> Real:
    """
    Calculates signed length of projection of one vector onto another.

    Positive sign of result means that angle between vectors is acute,
    negative -- obtuse,
    zero -- right.

    >>> signed_length((0, 0), (1, 0), (0, 0), (1, 0))
    1
    >>> signed_length((0, 0), (1, 0), (0, 0), (0, 1))
    0
    >>> signed_length((0, 0), (1, 0), (1, 0), (0, 0))
    -1
    """
    return parallelogram.signed_area(first_start, first_end,
                                     to_perpendicular_point(second_start),
                                     to_perpendicular_point(second_end))



from enum import (IntEnum,
                  unique)
from fractions import Fraction
from typing import Tuple

from .angular import (Orientation,
                      orientation)
from .hints import (Point,
                    Segment)
from .parallelogram import signed_area


@unique
class SegmentsRelationship(IntEnum):
    """
    Represents relationship between segments based on their intersection.
    """
    #: intersection is empty
    NONE = 0
    #: intersection is an endpoint of one of segments
    TOUCH = 1
    #: intersection is a point which is not an endpoint of any of segments
    CROSS = 2
    #: intersection is a segment itself
    OVERLAP = 3


def segments_relationship(left: Segment,
                          right: Segment) -> SegmentsRelationship:
    """
    Finds relationship between segments.

    >>> (segments_relationship(((0, 0), (2, 0)), ((0, 0), (2, 0)))
    ...  is SegmentsRelationship.OVERLAP)
    True
    >>> (segments_relationship(((0, 0), (2, 0)), ((0, 0), (0, 2)))
    ...  is SegmentsRelationship.TOUCH)
    True
    >>> (segments_relationship(((0, 0), (2, 0)), ((0, 0), (1, 0)))
    ...  is SegmentsRelationship.OVERLAP)
    True
    >>> (segments_relationship(((0, 0), (2, 0)), ((1, 0), (1, 1)))
    ...  is SegmentsRelationship.TOUCH)
    True
    >>> (segments_relationship(((0, 0), (2, 0)), ((1, 0), (2, 0)))
    ...  is SegmentsRelationship.OVERLAP)
    True
    >>> (segments_relationship(((0, 0), (2, 0)), ((2, 0), (3, 0)))
    ...  is SegmentsRelationship.TOUCH)
    True
    >>> (segments_relationship(((0, 0), (2, 0)), ((3, 0), (4, 0)))
    ...  is SegmentsRelationship.NONE)
    True
    """
    if left == right or left == right[::-1]:
        return SegmentsRelationship.OVERLAP
    left_start, left_end = left
    right_start, right_end = right
    if left_start > left_end:
        left_start, left_end = left_end, left_start
    if right_start > right_end:
        right_start, right_end = right_end, right_start
    left_start_orientation = orientation(right_start, right_end, left_start)
    left_end_orientation = orientation(right_start, right_end, left_end)
    if left_start_orientation is left_end_orientation:
        if left_start_orientation is Orientation.COLLINEAR:
            if left_start == right_start:
                return SegmentsRelationship.OVERLAP
            elif left_end == right_end:
                return SegmentsRelationship.OVERLAP
            elif left_start == right_end or left_end == right_start:
                return SegmentsRelationship.TOUCH
            elif right_start < left_start < right_end:
                return SegmentsRelationship.OVERLAP
            elif left_start < right_start < left_end:
                return SegmentsRelationship.OVERLAP
            else:
                return SegmentsRelationship.NONE
        else:
            return SegmentsRelationship.NONE
    elif left_start_orientation is Orientation.COLLINEAR:
        return (SegmentsRelationship.TOUCH
                if right_start <= left_start <= right_end
                else SegmentsRelationship.NONE)
    elif left_end_orientation is Orientation.COLLINEAR:
        return (SegmentsRelationship.TOUCH
                if right_start <= left_end <= right_end
                else SegmentsRelationship.NONE)
    else:
        right_start_orientation = orientation(left_end, left_start,
                                              right_start)
        right_end_orientation = orientation(left_end, left_start, right_end)
        if right_start_orientation is right_end_orientation:
            return SegmentsRelationship.NONE
        elif right_start_orientation is Orientation.COLLINEAR:
            return (SegmentsRelationship.TOUCH
                    if left_start < right_start < left_end
                    else SegmentsRelationship.NONE)
        elif right_end_orientation is Orientation.COLLINEAR:
            return (SegmentsRelationship.TOUCH
                    if left_start < right_end < left_end
                    else SegmentsRelationship.NONE)
        else:
            return SegmentsRelationship.CROSS


def segments_intersections(left: Segment, right: Segment) -> Tuple[Point, ...]:
    """
    Finds intersections of segments.

    >>> (segments_intersections(((0, 0), (2, 0)), ((0, 0), (2, 0)))
    ...  == ((0, 0), (2, 0)))
    True
    >>> segments_intersections(((0, 0), (2, 0)), ((0, 0), (0, 2))) == ((0, 0),)
    True
    >>> (segments_intersections(((0, 0), (2, 0)), ((0, 0), (1, 0)))
    ...  == ((0, 0), (1, 0)))
    True
    >>> segments_intersections(((0, 0), (2, 0)), ((1, 0), (1, 1))) == ((1, 0),)
    True
    >>> (segments_intersections(((0, 0), (2, 0)), ((1, 0), (2, 0)))
    ...  == ((1, 0), (2, 0)))
    True
    >>> segments_intersections(((0, 0), (2, 0)), ((2, 0), (3, 0))) == ((2, 0),)
    True
    >>> segments_intersections(((0, 0), (2, 0)), ((3, 0), (4, 0))) == ()
    True
    """
    relationship = segments_relationship(left, right)
    if relationship is SegmentsRelationship.NONE:
        return ()
    elif relationship is SegmentsRelationship.OVERLAP:
        _, first_intersection, second_intersection, _ = sorted(left + right)
        return first_intersection, second_intersection
    else:
        return segments_intersection(left, right),


def segments_intersection(left: Segment, right: Segment) -> Point:
    """
    Finds intersection point of segments that known to have only one.

    >>> segments_intersection(((0, 0), (2, 0)), ((0, 0), (0, 2))) == (0, 0)
    True
    >>> segments_intersection(((0, 0), (2, 0)), ((1, 0), (1, 1))) == (1, 0)
    True
    >>> segments_intersection(((0, 0), (2, 0)), ((2, 0), (3, 0))) == (2, 0)
    True
    """
    left_start, left_end = left
    right_start, right_end = right
    if segment_contains(left, right_start):
        return right_start
    elif segment_contains(left, right_end):
        return right_end
    elif segment_contains(right, left_start):
        return left_start
    elif segment_contains(right, left_end):
        return left_end
    else:
        denominator = signed_area(left_start, left_end, right_start, right_end)
        left_base_numerator = signed_area(left_start, right_start,
                                          right_start, right_end)
        right_base_numerator = signed_area(left_start, right_start,
                                           left_start, left_end)
        left_start_x, left_start_y = left_start
        left_end_x, left_end_y = left_end
        right_start_x, right_start_y = right_start
        right_end_x, right_end_y = right_end
        left_x_addend = (left_end_x - left_start_x) * left_base_numerator
        left_y_addend = (left_end_y - left_start_y) * left_base_numerator
        right_x_addend = (right_end_x - right_start_x) * right_base_numerator
        right_y_addend = (right_end_y - right_start_y) * right_base_numerator
        delta_x, delta_y = (abs(right_x_addend) - abs(left_x_addend),
                            abs(right_y_addend) - abs(left_y_addend))
        denominator_inv = (Fraction(1, denominator)
                           if isinstance(denominator, int)
                           else 1 / denominator)
        return (left_start_x + left_x_addend * denominator_inv
                if delta_x > 0
                else (right_start_x + right_x_addend * denominator_inv
                      if delta_x < 0
                      else (left_start_x + right_start_x
                            + (left_x_addend + right_x_addend)
                            * denominator_inv) / 2),
                left_start_y + left_y_addend * denominator_inv
                if delta_y > 0
                else (right_start_y + right_y_addend * denominator_inv
                      if delta_y < 0
                      else (left_start_y + right_start_y
                            + (left_y_addend + right_y_addend)
                            * denominator_inv) / 2))


def segment_contains(segment: Segment, point: Point) -> bool:
    """
    Checks if segment contains point.

    >>> segment_contains(((0, 0), (2, 0)), (0, 0))
    True
    >>> segment_contains(((0, 0), (2, 0)), (0, 2))
    False
    >>> segment_contains(((0, 0), (2, 0)), (1, 0))
    True
    >>> segment_contains(((0, 0), (2, 0)), (1, 1))
    False
    >>> segment_contains(((0, 0), (2, 0)), (2, 0))
    True
    >>> segment_contains(((0, 0), (2, 0)), (3, 0))
    False
    """
    start, end = segment
    return (point == start or point == end
            or (_bounding_box_contains(segment, point)
                and orientation(end, start, point) is Orientation.COLLINEAR))


def _bounding_box_contains(segment: Segment, point: Point) -> bool:
    (start_x, start_y), (end_x, end_y) = segment
    left_x, right_x = ((start_x, end_x)
                       if start_x < end_x
                       else (end_x, start_x))
    bottom_y, top_y = ((start_y, end_y)
                       if start_y < end_y
                       else (end_y, start_y))
    point_x, point_y = point
    return left_x <= point_x <= right_x and bottom_y <= point_y <= top_y



from numbers import Real
from typing import Tuple

from . import bounds
from .hints import (Expansion,
                    Point)


def fast_two_sum(left: Real, right: Real) -> Tuple[Real, Real]:
    estimation = left + right
    right_virtual = estimation - left
    tail = right - right_virtual
    return tail, estimation


def two_sum(left: Real, right: Real) -> Tuple[Real, Real]:
    estimation = left + right
    right_virtual = estimation - left
    left_virtual = estimation - right_virtual
    right_tail = right - right_virtual
    left_tail = left - left_virtual
    tail = left_tail + right_tail
    return tail, estimation


def split(value: Real,
          *,
          splitter: Real = bounds.splitter) -> Tuple[Real, Real]:
    base = splitter * value
    high = base - (base - value)
    low = value - high
    return low, high


def two_product_presplit(left: Real, right: Real, right_low: Real,
                         right_high: Real) -> Tuple[Real, Real]:
    estimation = left * right
    left_low, left_high = split(left)
    first_error = estimation - left_high * right_high
    second_error = first_error - left_low * right_high
    third_error = second_error - left_high * right_low
    tail = left_low * right_low - third_error
    return tail, estimation


def two_product(left: Real, right: Real) -> Tuple[Real, Real]:
    estimation = left * right
    left_low, left_high = split(left)
    right_low, right_high = split(right)
    first_error = estimation - left_high * right_high
    second_error = first_error - left_low * right_high
    third_error = second_error - left_high * right_low
    tail = left_low * right_low - third_error
    return tail, estimation


def two_two_diff(left_tail: Real, left: Real, right_tail: Real,
                 right: Real) -> Tuple[Real, Real, Real, Real]:
    third_tail, interim_tail, interim = two_one_diff(left_tail, left,
                                                     right_tail)
    second_tail, first_tail, estimation = two_one_diff(interim_tail, interim,
                                                       right)
    return third_tail, second_tail, first_tail, estimation


def two_two_sum(left_tail: Real, left: Real, right_tail: Real,
                right: Real) -> Tuple[Real, Real, Real, Real]:
    third_tail, interim_tail, interim = two_one_sum(left_tail, left,
                                                    right_tail)
    second_tail, first_tail, estimation = two_one_sum(interim_tail, interim,
                                                      right)
    return third_tail, second_tail, first_tail, estimation


def two_one_sum(left_tail: Real, left: Real,
                right: Real) -> Tuple[Real, Real, Real]:
    second_tail, interim = two_sum(left_tail, right)
    first_tail, estimation = two_sum(left, interim)
    return second_tail, first_tail, estimation


def two_one_diff(left_tail: Real, left: Real,
                 right: Real) -> Tuple[Real, Real, Real]:
    second_tail, interim = two_diff(left_tail, right)
    first_tail, estimation = two_sum(left, interim)
    return second_tail, first_tail, estimation


def two_diff(left: Real, right: Real) -> Tuple[Real, Real]:
    estimation = left - right
    return two_diff_tail(left, right, estimation), estimation


def two_diff_tail(left: Real, right: Real, estimation: Real) -> Real:
    right_virtual = left - estimation
    left_virtual = estimation + right_virtual
    right_error = right_virtual - right
    left_error = left - left_virtual
    return left_error + right_error


def square(value: Real) -> Tuple[Real, Real]:
    estimation = value * value
    value_low, value_high = split(value)
    first_error = estimation - value_high * value_high
    second_error = first_error - (value_high + value_high) * value_low
    tail = value_low * value_low - second_error
    return tail, estimation


def sum_expansions(left_expansion: Expansion,
                   right_expansion: Expansion) -> Expansion:
    """
    Sums two expansions with zero components elimination.
    """
    left_length, right_length = len(left_expansion), len(right_expansion)
    left_element, right_element = left_expansion[0], right_expansion[0]
    left_index = right_index = 0
    if (right_element > left_element) is (right_element > -left_element):
        accumulator = left_element
        left_index += 1
        try:
            left_element = left_expansion[left_index]
        except IndexError:
            pass
    else:
        accumulator = right_element
        right_index += 1
        try:
            right_element = right_expansion[right_index]
        except IndexError:
            pass
    result = []
    if (left_index < left_length) and (right_index < right_length):
        if (right_element > left_element) is (right_element > -left_element):
            tail, accumulator = fast_two_sum(left_element, accumulator)
            left_index += 1
            try:
                left_element = left_expansion[left_index]
            except IndexError:
                pass
        else:
            tail, accumulator = fast_two_sum(right_element, accumulator)
            right_index += 1
            try:
                right_element = right_expansion[right_index]
            except IndexError:
                pass
        if tail:
            result.append(tail)
        while (left_index < left_length) and (right_index < right_length):
            if ((right_element > left_element)
                    is (right_element > -left_element)):
                tail, accumulator = two_sum(accumulator, left_element)
                left_index += 1
                try:
                    left_element = left_expansion[left_index]
                except IndexError:
                    pass
            else:
                tail, accumulator = two_sum(accumulator, right_element)
                right_index += 1
                try:
                    right_element = right_expansion[right_index]
                except IndexError:
                    pass
            if tail:
                result.append(tail)
    while left_index < left_length:
        tail, accumulator = two_sum(accumulator, left_element)
        left_index += 1
        try:
            left_element = left_expansion[left_index]
        except IndexError:
            pass
        if tail:
            result.append(tail)
    while right_index < right_length:
        tail, accumulator = two_sum(accumulator, right_element)
        right_index += 1
        try:
            right_element = right_expansion[right_index]
        except IndexError:
            pass
        if tail:
            result.append(tail)
    if accumulator or not result:
        result.append(accumulator)
    return result


def scale_expansion(expansion: Expansion, scalar: Real) -> Expansion:
    """
    Multiplies an expansion by a scalar with zero components elimination.
    """
    expansion = iter(expansion)
    scalar_low, scalar_high = split(scalar)
    tail, accumulator = two_product_presplit(next(expansion), scalar,
                                             scalar_low, scalar_high)
    result = []
    if tail:
        result.append(tail)
    for element in expansion:
        product_tail, product = two_product_presplit(element, scalar,
                                                     scalar_low, scalar_high)
        tail, interim = two_sum(accumulator, product_tail)
        if tail:
            result.append(tail)
        tail, accumulator = fast_two_sum(product, interim)
        if tail:
            result.append(tail)
    if accumulator or not result:
        result.append(accumulator)
    return result


def to_cross_product(minuend_multiplier_x: Real,
                     minuend_multiplier_y: Real,
                     subtrahend_multiplier_x: Real,
                     subtrahend_multiplier_y: Real) -> Expansion:
    """
    Returns expansion of vectors' planar cross product.
    """
    minuend_tail, minuend = two_product(minuend_multiplier_x,
                                        minuend_multiplier_y)
    subtrahend_tail, subtrahend = two_product(subtrahend_multiplier_y,
                                              subtrahend_multiplier_x)
    return two_two_diff(minuend_tail, minuend, subtrahend_tail, subtrahend)


def to_perpendicular_point(point: Point) -> Point:
    x, y = point
    return -y, x


def to_sign(value: Real) -> int:
    if value > 0:
        return 1
    elif value < 0:
        return -1
    else:
        return 0



from numbers import Real
from typing import Tuple


def _to_epsilon_and_splitter() -> Tuple[Real, Real]:
    every_other = True
    epsilon, splitter = 1, 1
    check = 1
    while True:
        last_check = check
        epsilon /= 2
        if every_other:
            splitter *= 2
        every_other = not every_other
        check = 1 + epsilon
        if check == 1 or check == last_check:
            break
    splitter += 1
    return epsilon, splitter


epsilon, splitter = _to_epsilon_and_splitter()


def to_determinant_error(determinant: Real) -> Real:
    return (3 + 8 * epsilon) * epsilon * abs(determinant)


def to_signed_measure_first_error(upper_bound: Real) -> Real:
    return epsilon * (3 + 16 * epsilon) * upper_bound


def to_signed_measure_second_error(upper_bound: Real) -> Real:
    return epsilon * (2 + 12 * epsilon) * upper_bound


def to_signed_measure_third_error(upper_bound: Real) -> Real:
    return epsilon * epsilon * (9 + 64 * epsilon) * upper_bound


def to_cocircular_first_error(upper_bound: Real) -> Real:
    return (10 + 96 * epsilon) * epsilon * upper_bound


def to_cocircular_second_error(upper_bound: Real) -> Real:
    return epsilon * (4 + 48 * epsilon) * upper_bound


def to_cocircular_third_error(upper_bound: Real) -> Real:
    return epsilon * epsilon * (44 + 576 * epsilon) * upper_bound



from enum import (IntEnum,
                  unique)

from .hints import Point
from .parallelogram import signed_area
from .projection import signed_length
from .utils import to_sign


@unique
class Kind(IntEnum):
    """
    Represents kinds of angles
    based on their degrees value in range ``[0, 180]``.
    """
    #: ``(90, 180]`` degrees
    OBTUSE = -1
    #: ``90`` degrees
    RIGHT = 0
    #: ``[0, 90)`` degrees
    ACUTE = 1


@unique
class Orientation(IntEnum):
    """
    Represents kinds of angle orientations.
    """
    #: in the same direction as a clock's hands
    CLOCKWISE = -1
    #: to the top and then to the bottom or vice versa
    COLLINEAR = 0
    #: opposite to clockwise
    COUNTERCLOCKWISE = 1


def kind(first_ray_point: Point,
         vertex: Point,
         second_ray_point: Point) -> Kind:
    """
    Returns kind of angle built on given points.

    >>> kind((1, 0), (0, 0), (1, 0)) is Kind.ACUTE
    True
    >>> kind((1, 0), (0, 0), (0, 1)) is Kind.RIGHT
    True
    >>> kind((1, 0), (0, 0), (-1, 0)) is Kind.OBTUSE
    True
    """
    return Kind(to_sign(signed_length(vertex, first_ray_point,
                                      vertex, second_ray_point)))


def orientation(first_ray_point: Point,
                vertex: Point,
                second_ray_point: Point) -> Orientation:
    """
    Returns orientation of angle built on given points.

    >>> orientation((1, 0), (0, 0), (1, 0)) is Orientation.COLLINEAR
    True
    >>> orientation((1, 0), (0, 0), (0, 1)) is Orientation.COUNTERCLOCKWISE
    True
    >>> orientation((0, 1), (0, 0), (1, 0)) is Orientation.CLOCKWISE
    True
    """
    return Orientation(to_sign(signed_area(vertex, first_ray_point,
                                           vertex, second_ray_point)))



from functools import reduce
from numbers import Real
from typing import (Iterable,
                    Tuple)

from . import bounds
from .hints import (Expansion,
                    Point)
from .utils import (scale_expansion,
                    square,
                    sum_expansions,
                    to_cross_product,
                    two_diff_tail,
                    two_product,
                    two_two_diff,
                    two_two_sum)


def determinant(first_point: Point, second_point: Point,
                third_point: Point, fourth_point: Point) -> Real:
    """
    Calculates determinant of linear equations' system
    for checking if four points lie on the same circle.

    Positive sign of result means that point lies inside,
    negative -- outside,
    zero -- on a circle defined by other points.

    >>> determinant((0, 0), (2, 0), (2, 2), (0, 2))
    0
    >>> determinant((0, 0), (2, 0), (2, 2), (0, 3))
    -12
    >>> determinant((0, 0), (2, 0), (2, 2), (0, 1))
    4
    """
    first_x, first_y = first_point
    second_x, second_y = second_point
    third_x, third_y = third_point
    fourth_x, fourth_y = fourth_point

    first_dx, first_dy = first_x - fourth_x, first_y - fourth_y
    second_dx, second_dy = second_x - fourth_x, second_y - fourth_y
    third_dx, third_dy = third_x - fourth_x, third_y - fourth_y

    first_squared_distance = first_dx * first_dx + first_dy * first_dy
    second_squared_distance = second_dx * second_dx + second_dy * second_dy
    third_squared_distance = third_dx * third_dx + third_dy * third_dy

    first_dx_second_dy = first_dx * second_dy
    first_dx_third_dy = first_dx * third_dy
    second_dx_first_dy = second_dx * first_dy
    second_dx_third_dy = second_dx * third_dy
    third_dx_first_dy = third_dx * first_dy
    third_dx_second_dy = third_dx * second_dy

    result = (first_squared_distance
              * (second_dx_third_dy - third_dx_second_dy)
              + second_squared_distance
              * (third_dx_first_dy - first_dx_third_dy)
              + third_squared_distance
              * (first_dx_second_dy - second_dx_first_dy))
    upper_bound = (first_squared_distance
                   * (abs(second_dx_third_dy) + abs(third_dx_second_dy))
                   + second_squared_distance
                   * (abs(third_dx_first_dy) + abs(first_dx_third_dy))
                   + third_squared_distance
                   * (abs(first_dx_second_dy) + abs(second_dx_first_dy)))
    error_bound = bounds.to_cocircular_first_error(upper_bound)
    if result > error_bound or -result > error_bound:
        return result
    return _adjusted_determinant(first_x, first_y,
                                 second_x, second_y,
                                 third_x, third_y,
                                 fourth_x, fourth_y,
                                 upper_bound)


def _adjusted_determinant(first_x: Real, first_y: Real,
                          second_x: Real, second_y: Real,
                          third_x: Real, third_y: Real,
                          fourth_x: Real, fourth_y: Real,
                          upper_bound: Real) -> Real:
    first_dx, first_dy = first_x - fourth_x, first_y - fourth_y
    second_dx, second_dy = second_x - fourth_x, second_y - fourth_y
    third_dx, third_dy = third_x - fourth_x, third_y - fourth_y

    second_third_cross_product = to_cross_product(second_dx, third_dy,
                                                  third_dx, second_dy)
    third_first_cross_product = to_cross_product(third_dx, first_dy,
                                                 first_dx, third_dy)
    first_second_cross_product = to_cross_product(first_dx, second_dy,
                                                  second_dx, first_dy)
    result_expansion = sum_expansions(
            sum_expansions(
                    _multiply_by_squared_length(second_third_cross_product,
                                                first_dx, first_dy),
                    _multiply_by_squared_length(third_first_cross_product,
                                                second_dx, second_dy)),
            _multiply_by_squared_length(first_second_cross_product,
                                        third_dx, third_dy))
    result = sum(result_expansion)
    error_bound = bounds.to_cocircular_second_error(upper_bound)
    if result >= error_bound or -result >= error_bound:
        return result

    first_dx_tail = two_diff_tail(first_x, fourth_x, first_dx)
    first_dy_tail = two_diff_tail(first_y, fourth_y, first_dy)
    second_dx_tail = two_diff_tail(second_x, fourth_x, second_dx)
    second_dy_tail = two_diff_tail(second_y, fourth_y, second_dy)
    third_dx_tail = two_diff_tail(third_x, fourth_x, third_dx)
    third_dy_tail = two_diff_tail(third_y, fourth_y, third_dy)

    if (not first_dx_tail and not first_dy_tail
            and not second_dx_tail and not second_dy_tail
            and not third_dx_tail and not third_dy_tail):
        return result

    error_bound = (bounds.to_cocircular_third_error(upper_bound)
                   + bounds.to_determinant_error(result))
    result += (
            _to_addend(first_dx, first_dx_tail, first_dy, first_dy_tail,
                       second_dx, second_dx_tail, second_dy, second_dy_tail,
                       third_dx, third_dx_tail, third_dy, third_dy_tail)
            + _to_addend(second_dx, second_dx_tail, second_dy, second_dy_tail,
                         third_dx, third_dx_tail, third_dy, third_dy_tail,
                         first_dx, first_dx_tail, first_dy, first_dy_tail)
            + _to_addend(third_dx, third_dx_tail, third_dy, third_dy_tail,
                         first_dx, first_dx_tail, first_dy, first_dy_tail,
                         second_dx, second_dx_tail, second_dy, second_dy_tail))
    if result >= error_bound or -result >= error_bound:
        return result

    first_squared_length = (_to_squared_length(first_dx, first_dy)
                            if (second_dx_tail or second_dy_tail
                                or third_dx_tail or third_dy_tail)
                            else (0,) * 4)
    second_squared_length = (_to_squared_length(second_dx, second_dy)
                             if (first_dx_tail or first_dy_tail
                                 or third_dx_tail or third_dy_tail)
                             else (0,) * 4)
    third_squared_length = (_to_squared_length(third_dx, third_dy)
                            if (first_dx_tail or first_dy_tail
                                or second_dx_tail or second_dy_tail)
                            else (0,) * 4)

    if first_dx_tail:
        first_dx_tail_second_third_cross_product = scale_expansion(
                second_third_cross_product, first_dx_tail)
        result_expansion = sum_expansions(
                result_expansion,
                _to_extra(first_dx_tail_second_third_cross_product,
                          first_dx, first_dx_tail,
                          second_dy, second_squared_length,
                          third_dy, third_squared_length))

    if first_dy_tail:
        first_dy_tail_second_third_cross_product = scale_expansion(
                second_third_cross_product, first_dy_tail)
        result_expansion = sum_expansions(
                result_expansion,
                _to_extra(first_dy_tail_second_third_cross_product,
                          first_dy, first_dy_tail,
                          third_dx, third_squared_length,
                          second_dx, second_squared_length))

    if second_dx_tail:
        second_dx_tail_third_first_cross_product = scale_expansion(
                third_first_cross_product, second_dx_tail)
        result_expansion = sum_expansions(
                result_expansion,
                _to_extra(second_dx_tail_third_first_cross_product,
                          second_dx, second_dx_tail,
                          third_dy, third_squared_length,
                          first_dy, first_squared_length))

    if second_dy_tail:
        second_dy_tail_third_first_cross_product = scale_expansion(
                third_first_cross_product, second_dy_tail)
        result_expansion = sum_expansions(
                result_expansion,
                _to_extra(second_dy_tail_third_first_cross_product,
                          second_dy, second_dy_tail,
                          first_dx, first_squared_length,
                          third_dx, third_squared_length))

    if third_dx_tail:
        third_dx_tail_first_second_cross_product = scale_expansion(
                first_second_cross_product, third_dx_tail)
        result_expansion = sum_expansions(
                result_expansion,
                _to_extra(third_dx_tail_first_second_cross_product,
                          third_dx, third_dx_tail,
                          first_dy, first_squared_length,
                          second_dy, second_squared_length))

    if third_dy_tail:
        third_dy_tail_first_second_cross_product = scale_expansion(
                first_second_cross_product, third_dy_tail)
        result_expansion = sum_expansions(
                result_expansion,
                _to_extra(third_dy_tail_first_second_cross_product,
                          third_dy, third_dy_tail,
                          second_dx, second_squared_length,
                          first_dx, first_squared_length))

    if first_dx_tail or first_dy_tail:
        if second_dx_tail or second_dy_tail or third_dx_tail or third_dy_tail:
            second_third_crossed_tails_tail, second_third_crossed_tails = (
                _to_crossed_tails(second_dx, second_dx_tail,
                                  second_dy, second_dy_tail,
                                  third_dx, third_dx_tail,
                                  third_dy, third_dy_tail))
        else:
            second_third_crossed_tails_tail = second_third_crossed_tails = (0,)

        if first_dx_tail:
            result_expansion = reduce(
                    sum_expansions,
                    _to_dx_extras(first_dx_tail_second_third_cross_product,
                                  first_dx, first_dx_tail,
                                  second_dy_tail, second_squared_length,
                                  third_dy_tail, third_squared_length,
                                  second_third_crossed_tails,
                                  second_third_crossed_tails_tail),
                    result_expansion)

        if first_dy_tail:
            result_expansion = reduce(
                    sum_expansions,
                    _to_dy_extras(first_dy_tail_second_third_cross_product,
                                  first_dy, first_dy_tail,
                                  second_third_crossed_tails,
                                  second_third_crossed_tails_tail),
                    result_expansion)

    if second_dx_tail or second_dy_tail:
        if first_dx_tail or first_dy_tail or third_dx_tail or third_dy_tail:
            third_first_crossed_tails_tail, third_first_crossed_tails = (
                _to_crossed_tails(third_dx, third_dx_tail,
                                  third_dy, third_dy_tail,
                                  first_dx, first_dx_tail,
                                  first_dy, first_dy_tail))
        else:
            third_first_crossed_tails_tail = third_first_crossed_tails = (0,)

        if second_dx_tail:
            result_expansion = reduce(
                    sum_expansions,
                    _to_dx_extras(second_dx_tail_third_first_cross_product,
                                  second_dx, second_dx_tail,
                                  third_dy_tail, third_squared_length,
                                  first_dy_tail, first_squared_length,
                                  third_first_crossed_tails,
                                  third_first_crossed_tails_tail),
                    result_expansion)

        if second_dy_tail:
            result_expansion = reduce(
                    sum_expansions,
                    _to_dy_extras(second_dy_tail_third_first_cross_product,
                                  second_dy, second_dy_tail,
                                  third_first_crossed_tails,
                                  third_first_crossed_tails_tail),
                    result_expansion)

    if third_dx_tail or third_dy_tail:
        if first_dx_tail or first_dy_tail or second_dx_tail or second_dy_tail:
            first_second_crossed_tails_tail, first_second_crossed_tails = (
                _to_crossed_tails(first_dx, first_dx_tail,
                                  first_dy, first_dy_tail,
                                  second_dx, second_dx_tail,
                                  second_dy, second_dy_tail))
        else:
            first_second_crossed_tails_tail = first_second_crossed_tails = (0,)

        if third_dx_tail:
            result_expansion = reduce(
                    sum_expansions,
                    _to_dx_extras(third_dx_tail_first_second_cross_product,
                                  third_dx, third_dx_tail,
                                  first_dy_tail, first_squared_length,
                                  second_dy_tail, second_squared_length,
                                  first_second_crossed_tails,
                                  first_second_crossed_tails_tail),
                    result_expansion)

        if third_dy_tail:
            result_expansion = reduce(
                    sum_expansions,
                    _to_dy_extras(third_dy_tail_first_second_cross_product,
                                  third_dy, third_dy_tail,
                                  first_second_crossed_tails,
                                  first_second_crossed_tails_tail),
                    result_expansion)
    return result_expansion[-1]


def _to_dx_extras(expansion: Expansion,
                  dx: Real, dx_tail: Real,
                  left_dy_tail: Real, left_squared_length: Expansion,
                  right_dy_tail: Real, right_squared_length: Expansion,
                  left_right_crossed_tails: Expansion,
                  left_right_crossed_tails_tail: Expansion
                  ) -> Iterable[Expansion]:
    dx_tail_left_right_crossed_tails = scale_expansion(
            left_right_crossed_tails, dx_tail)
    yield sum_expansions(scale_expansion(expansion, dx_tail),
                         scale_expansion(dx_tail_left_right_crossed_tails,
                                         2 * dx))
    if left_dy_tail:
        yield scale_expansion(scale_expansion(right_squared_length, dx_tail),
                              left_dy_tail)
    if right_dy_tail:
        yield scale_expansion(scale_expansion(left_squared_length, -dx_tail),
                              right_dy_tail)
    first_addend = scale_expansion(dx_tail_left_right_crossed_tails, dx_tail)
    dx_tail_left_right_crossed_tails_tail = scale_expansion(
            left_right_crossed_tails_tail, dx_tail)
    second_addend = sum_expansions(
            scale_expansion(dx_tail_left_right_crossed_tails_tail, 2 * dx),
            scale_expansion(dx_tail_left_right_crossed_tails_tail, dx_tail))
    yield sum_expansions(first_addend, second_addend)


def _to_dy_extras(expansion: Expansion,
                  dy: Real, dy_tail: Real,
                  rest_crossed_tails: Expansion,
                  rest_crossed_tails_tail: Expansion) -> Iterable[Expansion]:
    dy_tail_rest_crossed_tails = scale_expansion(rest_crossed_tails, dy_tail)
    yield sum_expansions(scale_expansion(expansion, dy_tail),
                         scale_expansion(dy_tail_rest_crossed_tails, 2 * dy))
    first_addend = scale_expansion(dy_tail_rest_crossed_tails, dy_tail)
    dy_tail_rest_crossed_tails_tail = scale_expansion(rest_crossed_tails_tail,
                                                      dy_tail)
    second_addend = sum_expansions(
            scale_expansion(dy_tail_rest_crossed_tails_tail, 2 * dy),
            scale_expansion(dy_tail_rest_crossed_tails_tail, dy_tail))
    yield sum_expansions(first_addend, second_addend)


def _to_extra(expansion: Expansion,
              coordinate: Real, coordinate_tail: Real,
              left_coordinate: Real, left_squared_length: Expansion,
              right_coordinate: Real, right_squared_length: Expansion
              ) -> Expansion:
    second_addend = scale_expansion(scale_expansion(right_squared_length,
                                                    coordinate_tail),
                                    left_coordinate)
    first_addend = scale_expansion(expansion, 2 * coordinate)
    minuend = sum_expansions(first_addend, second_addend)
    subtrahend = scale_expansion(scale_expansion(left_squared_length,
                                                 coordinate_tail),
                                 -right_coordinate)
    return sum_expansions(subtrahend, minuend)


def _to_crossed_tails(left_dx: Real, left_dx_tail: Real,
                      left_dy: Real, left_dy_tail: Real,
                      right_dx: Real, right_dx_tail: Real,
                      right_dy: Real, right_dy_tail: Real
                      ) -> Tuple[Expansion, Expansion]:
    tail = two_two_diff(*two_product(left_dx_tail, right_dy_tail),
                        *two_product(right_dx_tail, left_dy_tail))
    estimation = sum_expansions(two_two_sum(*two_product(left_dx_tail,
                                                         right_dy),
                                            *two_product(left_dx,
                                                         right_dy_tail)),
                                two_two_sum(*two_product(right_dx_tail,
                                                         -left_dy),
                                            *two_product(right_dx,
                                                         -left_dy_tail)))
    return tail, estimation


def _multiply_by_squared_length(expansion: Expansion,
                                dx: Real, dy: Real) -> Expansion:
    return sum_expansions(scale_expansion(scale_expansion(expansion, dx), dx),
                          scale_expansion(scale_expansion(expansion, dy), dy))


def _to_addend(left_dx: Real, left_dx_tail: Real,
               left_dy: Real, left_dy_tail: Real,
               mid_dx: Real, mid_dx_tail: Real,
               mid_dy: Real, mid_dy_tail: Real,
               right_dx: Real, right_dx_tail: Real,
               right_dy: Real, right_dy_tail: Real) -> Real:
    return ((left_dx * left_dx + left_dy * left_dy)
            * ((mid_dx * right_dy_tail + right_dy * mid_dx_tail)
               - (mid_dy * right_dx_tail + right_dx * mid_dy_tail))
            + 2 * (left_dx * left_dx_tail + left_dy * left_dy_tail)
            * (mid_dx * right_dy - mid_dy * right_dx))


def _to_squared_length(dx: Real, dy: Real) -> Expansion:
    dx_squared_tail, dx_squared = square(dx)
    dy_squared_tail, dy_squared = square(dy)
    return two_two_sum(dx_squared_tail, dx_squared, dy_squared_tail,
                       dy_squared)
